BEGIN;
-- This file can only be used when 'table_version' is not
-- installed as an extension
DO $$
DECLARE
  cur_search_path text;
BEGIN
  IF EXISTS (
    SELECT * FROM pg_catalog.pg_extension WHERE extname = 'table_version'
  )
  THEN
    RAISE EXCEPTION 'table_version is installed as an extension';
  END IF;

  --SELECT reset_val INTO cur_search_path
  --FROM pg_settings WHERE name = 'search_path';
  EXECUTE 'SHOW search_path' INTO cur_search_path;
  EXECUTE 'SET search_path = table_version, ' || cur_search_path;

  IF NOT EXISTS (
    SELECT * FROM pg_catalog.pg_namespace WHERE nspname =
    '@extschema@'
  )
  THEN
    CREATE SCHEMA @extschema@;
  END IF;
END
$$ LANGUAGE 'plpgsql';
-- {
CREATE OR REPLACE FUNCTION _ver_get_table_info(regclass)
RETURNS TABLE (nspname name, relname name, rolname name) AS $$
    SELECT
      n.nspname, c.relname, r.rolname
    FROM
      pg_namespace n, pg_class c, pg_roles r
    WHERE
      c.oid = $1
    AND
      r.oid = c.relowner
    AND
      n.oid = c.relnamespace;
$$ LANGUAGE 'sql';
--}
--------------------------------------------------------------------------------

-- postgresql-table_version - PostgreSQL table versioning extension
--
-- Copyright 2016 Crown copyright (c)
-- Land Information New Zealand and the New Zealand Government.
-- All rights reserved
--
-- This software is released under the terms of the new BSD license. See the 
-- LICENSE file for more information.
--
--------------------------------------------------------------------------------

DO $$ BEGIN IF NOT EXISTS (
  SELECT t.oid
  FROM pg_class t, pg_namespace n
  WHERE n.oid = t.relnamespace
    AND n.nspname = '@extschema@'
    AND t.relname = 'revision'
) THEN
  CREATE TABLE IF NOT EXISTS revision (
      id SERIAL NOT NULL PRIMARY KEY,
      revision_time TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
      start_time TIMESTAMP NOT NULL DEFAULT clock_timestamp(),
      user_name TEXT NOT NULL DEFAULT CURRENT_USER,
      schema_change BOOLEAN NOT NULL,
      comment TEXT
  );
  RAISE NOTICE 'Setting revision_id_seq value to 1000';
  PERFORM setval('@extschema@.revision_id_seq', 1000, true);
END IF;
END; $$;

GRANT SELECT ON TABLE revision TO public;


COMMENT ON TABLE revision IS $$
Defines a revision represents a amendment to table or series of tables held 
within the database. Each revision is identified by an id.  

The revision_time is the datetime of the revision.

The start_time is the datetime of when the revision record was created.

The user_name is the database user who created the revision
$$;


CREATE TABLE IF NOT EXISTS versioned_tables (	
    id SERIAL NOT NULL PRIMARY KEY,
    schema_name NAME NOT NULL,
    table_name NAME NOT NULL,
    key_column VARCHAR(64) NOT NULL,
    versioned BOOLEAN NOT NULL,
    CONSTRAINT versioned_tables_name_key UNIQUE (schema_name, table_name)
);

GRANT SELECT ON TABLE versioned_tables TO public;

COMMENT ON TABLE versioned_tables IS $$
Defines if a table is versioned. Each table is identified by an id. 

The column used to define primary key
for the table is defined in key_column. This key does not actually have to be
table primary key, rather itneeds to be a unique non-composite integer or bigint
column.
$$;


CREATE TABLE IF NOT EXISTS tables_changed (
    revision INTEGER NOT NULL REFERENCES revision,
    table_id INTEGER NOT NULL REFERENCES versioned_tables,
    PRIMARY KEY (revision, table_id)
);

GRANT SELECT ON TABLE tables_changed TO public;

COMMENT ON TABLE tables_changed IS $$
Defines which tables are modified by a given revision.
$$;




-- {
CREATE OR REPLACE FUNCTION ver_enable_versioning(
    v_table_oid       REGCLASS
) 
RETURNS BOOLEAN AS
$$
DECLARE
    v_schema          NAME;
    v_table           NAME;
    v_owner           NAME;
    v_key_col         NAME;
    v_revision_table  TEXT;
    v_sql             TEXT;
    v_table_id        @extschema@.versioned_tables.id%TYPE;
    v_revision        @extschema@.revision.id%TYPE;
    v_revision_exists BOOLEAN;
    v_table_has_data  BOOLEAN;
    v_role            TEXT;
    v_privilege       TEXT;
BEGIN

    SELECT nspname, relname, rolname
    INTO v_schema, v_table, v_owner
    FROM @extschema@._ver_get_table_info(v_table_oid);

    -- Check that SESSION_USER is the owner of the table, or
    -- refuse to enable versioning on this table
    IF NOT pg_has_role(session_user, v_owner, 'usage') THEN
        RAISE EXCEPTION 'User % cannot enable versioning on table %'
            ' for lack of usage privileges on table owner role %',
            session_user, v_table_oid, v_owner;
    END IF;

    SELECT
        ATT.attname as col
    INTO
        v_key_col
    FROM
        pg_index IDX,
        pg_attribute ATT
    WHERE
        IDX.indrelid = v_table_oid AND
        ATT.attrelid = v_table_oid AND
        ATT.attnum = ANY(IDX.indkey) AND
        ATT.attnotnull = TRUE AND
        IDX.indisunique = TRUE AND
        IDX.indexprs IS NULL AND
        IDX.indpred IS NULL AND
        format_type(ATT.atttypid, NULL) IN ('integer', 'bigint', 'text', 'character varying') AND
        array_length(IDX.indkey::INTEGER[], 1) = 1
    ORDER BY
        IDX.indisprimary DESC
    LIMIT 1;

    IF v_key_col IS NULL THEN
        RAISE EXCEPTION 'Table % does not have a unique non-compostite integer, bigint, text, or varchar column', v_table_oid::text;
    END IF;

    IF (SELECT count(*) <= 1 FROM information_schema.columns WHERE table_name= v_table AND table_schema = v_schema) THEN
        RAISE EXCEPTION 'Table % must contain at least one other non key column', v_table_oid::text;
    END IF;

    v_revision_table := @extschema@.ver_get_version_table_full(v_schema, v_table);
    
    v_sql :=
    'CREATE TABLE ' || v_revision_table || '(' ||
        '_revision_created INTEGER NOT NULL,' ||
        '_revision_expired INTEGER,' ||
        'LIKE ' || v_table_oid::text ||
    ');';
    BEGIN
      EXECUTE v_sql;
    EXCEPTION
    WHEN duplicate_table THEN
        RAISE EXCEPTION 'Table %.% is already versioned', quote_ident(v_schema), quote_ident(v_table);
    END;
    
    v_sql := 'ALTER TABLE ' || v_revision_table || ' OWNER TO ' || 
        @extschema@._ver_get_table_owner(v_table_oid);
    EXECUTE v_sql;
    
    -- replicate permissions from source table to revision history table
    FOR v_role, v_privilege IN
        SELECT CASE WHEN g.grantee = 'PUBLIC'
                   THEN 'public'
                   ELSE g.grantee
               END as grantee,
               g.privilege_type
        FROM information_schema.role_table_grants g
        WHERE g.table_name = v_table
        AND   g.table_schema =  v_schema
    LOOP
        EXECUTE 'GRANT ' || v_privilege || ' ON TABLE ' || v_revision_table || 
            ' TO ' || quote_ident(v_role);
    END LOOP;
        
    v_sql := (
        SELECT
            'ALTER TABLE  ' || v_revision_table || ' ALTER COLUMN ' || attname || ' SET STATISTICS ' ||  attstattarget
        FROM
            pg_attribute 
        WHERE
            attrelid = v_table_oid AND
            attname = v_key_col AND
            attisdropped IS FALSE AND
            attnum > 0 AND
            attstattarget > 0
    );
    IF v_sql IS NOT NULL THEN
        EXECUTE v_sql;
    END IF;
    
    -- insert base data into table using a revision that is currently in
    -- progress, or if one does not exist create one.
    
    v_revision_exists := FALSE;
    
    EXECUTE 'SELECT EXISTS (SELECT * FROM ' || CAST(v_table_oid AS TEXT) || ' LIMIT 1)'
    INTO v_table_has_data;
    
    IF v_table_has_data THEN
        IF @extschema@._ver_get_reversion_temp_table('_changeset_revision') THEN
            SELECT
                max(VER.revision)
            INTO
                v_revision
            FROM
                _changeset_revision VER;
            
            v_revision_exists := TRUE;
        ELSE
            SELECT @extschema@.ver_create_revision(
                'Initial revisioning of ' || CAST(v_table_oid AS TEXT)
            )
            INTO  v_revision;
        END IF;
    
        v_sql :=
            'INSERT INTO ' || v_revision_table ||
            ' SELECT ' || v_revision || ', NULL, * FROM ' || CAST(v_table_oid AS TEXT);
        EXECUTE v_sql;
        
        IF NOT v_revision_exists THEN
            PERFORM @extschema@.ver_complete_revision();
        END IF;
    
    END IF;

    v_sql := 'ALTER TABLE  ' || v_revision_table || ' ADD CONSTRAINT ' ||
        quote_ident('pkey_' || v_revision_table) || ' PRIMARY KEY(_revision_created, ' ||
        quote_ident(v_key_col) || ')';
    EXECUTE v_sql;
    
    v_sql := 'CREATE INDEX ' || quote_ident('idx_' || v_table) || '_' || quote_ident(v_key_col) || ' ON ' || v_revision_table ||
        '(' || quote_ident(v_key_col) || ')';
    EXECUTE v_sql;

    v_sql := 'CREATE INDEX ' || quote_ident('fk_' || v_table) || '_expired ON ' || v_revision_table ||
        '(_revision_expired)';
    EXECUTE v_sql;

    v_sql := 'CREATE INDEX ' || quote_ident('fk_' || v_table) || '_created ON ' || v_revision_table ||
        '(_revision_created)';
    EXECUTE v_sql;
    
    EXECUTE 'ANALYSE ' || v_revision_table;

    SELECT
        id
    INTO
        v_table_id
    FROM
        @extschema@.versioned_tables
    WHERE
        schema_name = v_schema AND
        table_name = v_table;

    IF v_table_id IS NOT NULL THEN
        UPDATE @extschema@.versioned_tables
        SET    versioned = TRUE
        WHERE  schema_name = v_schema
        AND    table_name = v_table;
    ELSE
        INSERT INTO @extschema@.versioned_tables(schema_name, table_name, key_column, versioned)
        VALUES (v_schema, v_table, v_key_col, TRUE)
        RETURNING id INTO v_table_id;
    END IF;
    
    IF v_table_id IS NOT NULL AND v_table_has_data THEN
        INSERT INTO @extschema@.tables_changed(
            revision,
            table_id
        )
        SELECT
            v_revision,
            v_table_id
        WHERE
            NOT EXISTS (
                SELECT *
                FROM   @extschema@.tables_changed
                WHERE  table_id = v_table_id
                AND    revision = v_revision
        );
    END IF;

    PERFORM @extschema@.ver_create_table_functions(v_schema, v_table, v_key_col);
    PERFORM @extschema@.ver_create_version_trigger(v_table_oid, v_key_col);
    
    RETURN TRUE;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;
--}

CREATE OR REPLACE FUNCTION ver_enable_versioning(p_schema NAME, p_table  NAME)
RETURNS BOOLEAN AS $$
  SELECT @extschema@.ver_enable_versioning(( p_schema || '.' || p_table)::regclass);
$$ LANGUAGE sql;

-- Abort drop of versioned tables.
CREATE OR REPLACE FUNCTION _ver_abort_drop_of_versioned_table()
RETURNS event_trigger AS $$
DECLARE
    obj RECORD;
BEGIN

    FOR obj IN SELECT *
               FROM pg_event_trigger_dropped_objects()
               WHERE object_type = 'table'
    LOOP
        IF @extschema@._ver_is_table_versioned(obj.schema_name, obj.object_name)
        THEN
            RAISE EXCEPTION
                'cannot drop versioned table %, unversion it first',
                obj.object_identity;
        END IF;
    END LOOP;
END;
$$ LANGUAGE plpgsql;

-- NOTE: we need the DROP EVENT because CREATE OR REPLACE does not
--       exist for events. We _assume_ this script is being loaded
--       within a transaction (as done by table_version-loader)
--       so that the trigger will always be in effect.
DROP EVENT TRIGGER IF EXISTS _ver_abort_drop_of_versioned_table;
CREATE EVENT TRIGGER _ver_abort_drop_of_versioned_table
ON sql_drop EXECUTE PROCEDURE _ver_abort_drop_of_versioned_table();
-- {
CREATE OR REPLACE FUNCTION ver_disable_versioning(
    p_table_oid       REGCLASS
) 
RETURNS BOOLEAN AS
$$
DECLARE
    v_schema NAME;
    v_table  NAME;
    v_owner  NAME;
BEGIN

    SELECT nspname, relname, rolname
    INTO v_schema, v_table, v_owner
    FROM @extschema@._ver_get_table_info(p_table_oid);

    -- Check that SESSION_USER is the owner of the table, or
    -- refuse to enable versioning on this table
    IF NOT pg_has_role(session_user, v_owner, 'usage') THEN
        RAISE EXCEPTION 'User % cannot disable versioning on table %'
            ' for lack of usage privileges on table owner role %',
            session_user, p_table_oid, v_owner;
    END IF;

    IF NOT (SELECT @extschema@.ver_is_table_versioned(v_schema, v_table)) THEN
        RAISE EXCEPTION 'Table %.% is not versioned', quote_ident(v_schema), quote_ident(v_table);
    END IF;
    
    UPDATE @extschema@.versioned_tables
    SET    versioned = FALSE
    WHERE  schema_name = v_schema
    AND    table_name = v_table;

    EXECUTE 'DROP TRIGGER IF EXISTS '  || @extschema@._ver_get_version_trigger(v_schema, v_table) || ' ON ' ||  
        quote_ident(v_schema) || '.' || quote_ident(v_table);
    EXECUTE 'DROP FUNCTION IF EXISTS ' || @extschema@.ver_get_version_table_full(v_schema, v_table) || '()';
    EXECUTE 'DROP FUNCTION IF EXISTS ' || @extschema@._ver_get_diff_function(v_schema, v_table);
    EXECUTE 'DROP FUNCTION IF EXISTS ' || @extschema@._ver_get_revision_function(v_schema, v_table);
    EXECUTE 'DROP TABLE IF EXISTS '    || @extschema@.ver_get_version_table_full(v_schema, v_table) || ' CASCADE';    

    EXECUTE 'WITH deleted AS ('
            ' DELETE FROM table_version.versioned_tables'
            ' WHERE schema_name=$1'
            ' AND table_name=$2 RETURNING id'
            ') DELETE FROM table_version.tables_changed'
            '  WHERE table_id in ( select * from deleted )'
    USING v_schema, v_table;
    
    RETURN TRUE;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;
--}

CREATE OR REPLACE FUNCTION ver_disable_versioning(p_schema NAME, p_table  NAME)
RETURNS BOOLEAN AS $$
  SELECT @extschema@.ver_disable_versioning(( p_schema || '.' || p_table)::regclass);
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_create_revision(
    p_comment       TEXT, 
    p_revision_time TIMESTAMP DEFAULT CURRENT_TIMESTAMP, 
    p_schema_change BOOLEAN DEFAULT FALSE
) 
RETURNS INTEGER AS
$$
DECLARE
    v_revision @extschema@.revision.id%TYPE;
BEGIN
    IF @extschema@._ver_get_reversion_temp_table('_changeset_revision') THEN
        RAISE EXCEPTION 'A revision changeset is still in progress. Please complete the revision before starting a new one';
    END IF;

    INSERT INTO @extschema@.revision (revision_time, schema_change, comment, user_name)
    VALUES (p_revision_time, p_schema_change, p_comment, SESSION_USER)
    RETURNING id INTO v_revision;
    
    CREATE TEMP TABLE _changeset_revision(
        revision INTEGER NOT NULL PRIMARY KEY
    );
    INSERT INTO _changeset_revision(revision) VALUES (v_revision);
    ANALYSE _changeset_revision;
    
    RETURN v_revision;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;

CREATE OR REPLACE FUNCTION ver_complete_revision() RETURNS BOOLEAN AS
$$
BEGIN
    IF NOT @extschema@._ver_get_reversion_temp_table('_changeset_revision') THEN
        RETURN FALSE;
    END IF;
    
    DROP TABLE _changeset_revision;
    
    RETURN TRUE;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;


CREATE OR REPLACE FUNCTION ver_delete_revision(
    p_revision INTEGER
) 
RETURNS BOOLEAN AS
$$
DECLARE
    v_status BOOLEAN;
BEGIN
    BEGIN
        DELETE FROM @extschema@.revision
        WHERE id = p_revision;
        v_status := FOUND;
    EXCEPTION
        WHEN foreign_key_violation THEN
            RAISE WARNING 'Can not delete revision % as it is referenced by other tables: %', p_revision, SQLERRM;
            v_status := FALSE;
    END;
    RETURN v_status;
END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION ver_get_revision(
    p_revision        INTEGER, 
    OUT id            INTEGER, 
    OUT revision_time TIMESTAMP,
    OUT start_time    TIMESTAMP,
    OUT schema_change BOOLEAN,
    OUT comment       TEXT,
    OUT user_name     TEXT
) AS $$
    SELECT
        id,
        revision_time,
        start_time,
        schema_change,
        comment,
        user_name
    FROM
        @extschema@.revision
    WHERE
        id = $1
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_get_revisions(p_revisions INTEGER[]) 
RETURNS TABLE(
    id             INTEGER,
    revision_time  TIMESTAMP,
    start_time     TIMESTAMP,
    schema_change  BOOLEAN,
    comment        TEXT,
    user_name      TEXT
) AS $$
    SELECT
        id,
        revision_time,
        start_time,
        schema_change,
        comment,
        user_name
    FROM
        @extschema@.revision
    WHERE
        id = ANY($1)
    ORDER BY
        revision DESC;
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_get_revisions(
    p_start_date TIMESTAMP,
    p_end_date   TIMESTAMP
)
RETURNS TABLE(
    id             INTEGER
) AS $$
    SELECT
        id
    FROM
        @extschema@.revision
    WHERE
        revision_time >= $1 AND
        revision_time <= $2
    ORDER BY
        revision DESC;
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_get_revision(
    p_date_time       TIMESTAMP
) 
RETURNS INTEGER AS 
$$
    SELECT
        id
    FROM
        @extschema@.revision
    WHERE
        id IN (
            SELECT max(id) 
            FROM   @extschema@.revision
            WHERE  revision_time <= $1
        );
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_get_last_revision() 
RETURNS INTEGER AS 
$$
    SELECT
        id
    FROM
        @extschema@.revision
    WHERE
        id IN (
            SELECT max(id) 
            FROM   @extschema@.revision
        );
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_get_table_base_revision(
    p_schema          NAME,
    p_table           NAME
)
RETURNS INTEGER AS
$$
DECLARE
    v_id INTEGER;
BEGIN
    IF NOT @extschema@.ver_is_table_versioned(p_schema, p_table) THEN
        RAISE EXCEPTION 'Table %.% is not versioned', quote_ident(p_schema), quote_ident(p_table);
    END IF;
    
    SELECT
        VER.id
    INTO
        v_id
    FROM
        @extschema@.revision VER
    WHERE
        VER.id IN (
            SELECT min(TBC.revision)
            FROM   @extschema@.versioned_tables VTB,
                   @extschema@.tables_changed TBC
            WHERE  VTB.schema_name = p_schema
            AND    VTB.table_name = p_table
            AND    VTB.id = TBC.table_id
        );
    RETURN v_id;
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION ver_get_table_last_revision(
    p_schema          NAME,
    p_table           NAME
) 
RETURNS INTEGER AS
$$
DECLARE
    v_id INTEGER;
BEGIN
    IF NOT @extschema@.ver_is_table_versioned(p_schema, p_table) THEN
        RAISE EXCEPTION 'Table %.% is not versioned', quote_ident(p_schema), quote_ident(p_table);
    END IF;
    
    SELECT
        VER.id
    INTO
        v_id
    FROM
        @extschema@.revision VER
    WHERE
        VER.id IN (
            SELECT max(TBC.revision)
            FROM   @extschema@.versioned_tables VTB,
                   @extschema@.tables_changed TBC
            WHERE  VTB.schema_name = p_schema
            AND    VTB.table_name = p_table
            AND    VTB.id = TBC.table_id
        );
    RETURN v_id;
END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION ver_get_modified_tables(
    p_revision  INTEGER
)
RETURNS TABLE(
    schema_name NAME,
    table_name  NAME
) 
AS $$
BEGIN
    IF NOT EXISTS(SELECT * FROM @extschema@.revision WHERE id = p_revision) THEN
        RAISE EXCEPTION 'Revision % does not exist', p_revision;
    END IF;
            
    RETURN QUERY
        SELECT
            VTB.schema_name,
            VTB.table_name
        FROM
            @extschema@.versioned_tables VTB,
            @extschema@.tables_changed TBC
        WHERE
            VTB.id = TBC.table_id AND
            TBC.revision = p_revision
        ORDER BY
            VTB.schema_name,
            VTB.table_name;
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION ver_get_modified_tables(
    p_revision1 INTEGER,
    p_revision2 INTEGER
) 
RETURNS TABLE(
    revision    INTEGER,
    schema_name NAME,
    table_name  NAME
) AS
$$
DECLARE
    v_revision1 INTEGER;
    v_revision2 INTEGER;
    v_temp      INTEGER;
BEGIN
    v_revision1 := p_revision1;
    v_revision2 := p_revision2;

    IF v_revision1 > v_revision2 THEN
        v_temp      := v_revision1;
        v_revision1 := v_revision2;
        v_revision2 := v_temp;
    END IF;
    
    RETURN QUERY
        SELECT
            TBC.revision,
            VTB.schema_name,
            VTB.table_name
        FROM
            @extschema@.versioned_tables VTB,
            @extschema@.tables_changed TBC
        WHERE
            VTB.id = TBC.table_id AND
            TBC.revision > v_revision1 AND
            TBC.revision <= v_revision2
        ORDER BY
            TBC.revision,
            VTB.schema_name,
            VTB.table_name;
END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION _ver_is_table_versioned(
    p_schema NAME,
    p_table  NAME
)
RETURNS BOOLEAN AS
$$
    SELECT EXISTS (
      SELECT
          versioned
      FROM
          @extschema@.versioned_tables
      WHERE
          schema_name = p_schema AND
          table_name = p_table
    )
$$ LANGUAGE sql STABLE;

-- {
CREATE OR REPLACE FUNCTION ver_is_table_versioned(
    p_schema NAME,
    p_table  NAME
)
RETURNS BOOLEAN AS
$$
DECLARE
    v_is_versioned BOOLEAN;
BEGIN
    v_is_versioned := @extschema@._ver_is_table_versioned(p_schema, p_table);

    IF v_is_versioned IS FALSE THEN
        RETURN FALSE;
    END IF;

    -- Check that table exists and triggers are in place;
    -- warn otherwise, see:
    -- https://github.com/linz/postgresql-tableversion/issues/57
    BEGIN
      v_is_versioned := EXISTS (
        SELECT *
      FROM pg_trigger t
      WHERE t.tgrelid = (quote_ident(p_schema) || '.' || quote_ident(p_table) )::regclass
        AND t.tgname = p_schema || '_' || p_table || '_revision_trg'
      );
    EXCEPTION
      WHEN UNDEFINED_TABLE THEN
        RAISE WARNING 'Table %.% does not exist', p_schema, p_table;
        RETURN FALSE;
      WHEN OTHERS THEN
        RAISE EXCEPTION 'Got % (%)', SQLERRM, SQLSTATE;
    END;

    IF NOT v_is_versioned THEN
        RAISE WARNING 'Table %.% is known as versioned but lacks versioning triggers',
          p_schema, p_table
        USING HINT = 'use ver_enable_versioning to recover';
    END IF;

    RETURN v_is_versioned;
END;
$$ LANGUAGE plpgsql STABLE;
-- }


-- {
CREATE OR REPLACE FUNCTION ver_versioned_table_change_column_type(
    p_table_oid   REGCLASS,
    p_column_name NAME,
    p_column_datatype TEXT
)
RETURNS boolean AS
$$
DECLARE
    v_key_col NAME;
    v_trigger_name TEXT;
    v_revision_table TEXT;
    v_schema          NAME;
    v_table           NAME;
    v_owner           NAME;
BEGIN

    SELECT nspname, relname, rolname
    INTO v_schema, v_table, v_owner
    FROM @extschema@._ver_get_table_info(p_table_oid);

    -- Check that SESSION_USER is the owner of the table, or
    -- refuse to add columns to it
    IF NOT pg_has_role(session_user, v_owner, 'usage') THEN
        RAISE EXCEPTION 'User % cannot change type of columns on table %'
            ' for lack of usage privileges on table owner role %',
            session_user, p_table_oid, v_owner;
    END IF;

    IF NOT @extschema@.ver_is_table_versioned(v_schema, v_table) THEN
        RAISE EXCEPTION 'Table %.% is not versioned', quote_ident(v_schema), quote_ident(v_table);
    END IF;
    
    v_revision_table := @extschema@.ver_get_version_table_full(v_schema, v_table);
    v_trigger_name := @extschema@._ver_get_version_trigger(v_schema, v_table);
    
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' ||
        quote_ident(v_table) || ' DISABLE TRIGGER ' || v_trigger_name;
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' ||
        quote_ident(v_table) || ' ALTER COLUMN ' || quote_ident(p_column_name) ||
        ' TYPE ' || p_column_datatype;
    EXECUTE 'ALTER TABLE ' || v_revision_table      || ' ALTER COLUMN ' ||
        quote_ident(p_column_name) || ' TYPE ' || p_column_datatype;
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' ||
        quote_ident(v_table) || ' ENABLE TRIGGER ' || v_trigger_name;
    
    SELECT
        key_column
    INTO
        v_key_col
    FROM 
        @extschema@.versioned_tables
    WHERE
        schema_name = v_schema AND
        table_name = v_table;
    
    PERFORM @extschema@.ver_create_table_functions(v_schema, v_table, v_key_col);
    PERFORM @extschema@.ver_create_version_trigger(p_table_oid, v_key_col);
    RETURN TRUE;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;
--}

CREATE OR REPLACE FUNCTION ver_versioned_table_change_column_type(
    p_schema_name NAME,
    p_table_name  NAME,
    p_column_name NAME,
    p_column_datatype TEXT
)
RETURNS boolean AS
$$
    SELECT @extschema@.ver_versioned_table_change_column_type(
        ( p_schema_name || '.' || p_table_name)::regclass,
        p_column_name,
        p_column_datatype
    );
$$ LANGUAGE sql;


-- {
CREATE OR REPLACE FUNCTION ver_versioned_table_add_column(
    p_table_oid   REGCLASS,
    p_column_name NAME,
    p_column_datatype TEXT
)
RETURNS BOOLEAN AS 
$$
DECLARE
    v_key_col NAME;
    v_trigger_name TEXT;
    v_revision_table TEXT;
    v_schema          NAME;
    v_table           NAME;
    v_owner           NAME;
BEGIN

    SELECT nspname, relname, rolname
    INTO v_schema, v_table, v_owner
    FROM @extschema@._ver_get_table_info(p_table_oid);

    -- Check that SESSION_USER is the owner of the table, or
    -- refuse to add columns to it
    IF NOT pg_has_role(session_user, v_owner, 'usage') THEN
        RAISE EXCEPTION 'User % cannot add columns on table %'
            ' for lack of usage privileges on table owner role %',
            session_user, p_table_oid, v_owner;
    END IF;
    
    IF NOT @extschema@.ver_is_table_versioned(v_schema, v_table) THEN
        RAISE EXCEPTION 'Table %.% is not versioned',
            quote_ident(v_schema), quote_ident(v_table);
    END IF;

    v_revision_table := @extschema@.ver_get_version_table_full(v_schema, v_table);
    v_trigger_name := @extschema@._ver_get_version_trigger(v_schema, v_table);
    
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' || 
        quote_ident(v_table) || ' DISABLE TRIGGER ' || v_trigger_name;
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' ||
        quote_ident(v_table) || ' ADD COLUMN ' ||
        quote_ident(p_column_name) || ' ' || p_column_datatype;
    EXECUTE 'ALTER TABLE ' || v_revision_table      || ' ADD COLUMN ' ||
        quote_ident(p_column_name) || ' ' || p_column_datatype;
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' ||
        quote_ident(v_table) || ' ENABLE TRIGGER ' || v_trigger_name;
    
    SELECT
        key_column
    INTO
        v_key_col
    FROM 
        @extschema@.versioned_tables
    WHERE
        schema_name = v_schema AND
        table_name = v_table;
    
    PERFORM @extschema@.ver_create_table_functions(v_schema, v_table, v_key_col);
    PERFORM @extschema@.ver_create_version_trigger(p_table_oid, v_key_col);
    RETURN TRUE;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;
--}

CREATE OR REPLACE FUNCTION ver_versioned_table_add_column(
    p_schema_name NAME,
    p_table_name  NAME,
    p_column_name NAME,
    p_column_datatype TEXT
)
RETURNS BOOLEAN AS $$
    SELECT @extschema@.ver_versioned_table_add_column(
        ( p_schema_name || '.' || p_table_name)::regclass,
        p_column_name,
        p_column_datatype
    );
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_get_versioned_tables()
RETURNS TABLE(
    schema_name NAME,
    table_name  NAME,
    key_column  VARCHAR(64)
) AS $$
    SELECT
        schema_name,
        table_name,
        key_column
    FROM 
        @extschema@.versioned_tables
    WHERE
        versioned = TRUE AND
        @extschema@.ver_is_table_versioned(schema_name, table_name);
$$ LANGUAGE sql;

/**
* Get the versioned table key
*
* @return       The versioned table key
*/
CREATE OR REPLACE FUNCTION ver_get_versioned_table_key(
    p_schema_name NAME,
    p_table_name  NAME
)
RETURNS VARCHAR(64)
AS $$
    SELECT
        key_column
    FROM 
        @extschema@.versioned_tables
    WHERE
        versioned = TRUE AND
        schema_name = $1 AND
        table_name = $2 AND
        @extschema@.ver_is_table_versioned(schema_name, table_name);
$$ LANGUAGE sql;


/**
* Creates functions required for versioning the table.
*
* @param p_schema         The table schema
* @param p_table          The table name
* @param p_key_col        The unique non-compostite integer column key.
* @return                 If creating the functions was successful.
* @throws RAISE_EXCEPTION If the table is not versioned
* @throws RAISE_EXCEPTION If the table column definition could not be found
*/
CREATE OR REPLACE FUNCTION ver_create_table_functions(
    p_schema  NAME, 
    p_table   NAME, 
    p_key_col NAME
) 
RETURNS BOOLEAN AS
$$
DECLARE
    v_revision_table      TEXT;
    v_sql                 TEXT;
    v_col_cur             refcursor;
    v_column_name         NAME;
    v_column_type         TEXT;
    v_table_columns       TEXT;
    v_select_columns_diff TEXT;
    v_select_columns_rev  TEXT;
BEGIN
    IF NOT @extschema@._ver_is_table_versioned(p_schema, p_table) THEN
        RAISE EXCEPTION 'Table %.% is not versioned', quote_ident(p_schema), quote_ident(p_table);
    END IF;
    
    v_revision_table := @extschema@.ver_get_version_table_full(p_schema, p_table);
    v_table_columns := '';
    v_select_columns_diff := '';
    v_select_columns_rev := '';
    
    OPEN v_col_cur FOR
    SELECT att_name AS column_name, att_type AS column_type
    FROM unnest(@extschema@._ver_get_table_columns(p_schema || '.' ||  p_table));

    FETCH FIRST IN v_col_cur INTO v_column_name, v_column_type;
    LOOP
        v_select_columns_rev := v_select_columns_rev || REPEAT(' ', 16) || 'T.' || quote_ident(v_column_name);
        v_select_columns_diff := v_select_columns_diff || REPEAT(' ', 16) || 'TL.' || quote_ident(v_column_name);
        v_table_columns := v_table_columns || '    ' || quote_ident(v_column_name) || ' ' || v_column_type;
        FETCH v_col_cur INTO v_column_name, v_column_type;
        IF FOUND THEN
            v_select_columns_rev :=  v_select_columns_rev || ', ' || E'\n';
            v_select_columns_diff :=  v_select_columns_diff || ', ' || E'\n';
            v_table_columns :=   v_table_columns  || ', ' || E'\n';
        ELSE
            v_table_columns  :=  v_table_columns  || E'\n';
            EXIT;
        END IF;
    END LOOP;
    
    CLOSE v_col_cur;

    -- Create difference function for table called:
    -- ver_get_$schema$_$table$_diff(p_revision1 integer, p_revision2 integer)
    v_sql := $template$
    
CREATE OR REPLACE FUNCTION %func_sig%
RETURNS TABLE(
    _diff_action CHAR(1),
    %table_columns%
) 
AS $FUNC$
    DECLARE
        v_revision1      INTEGER;
        v_revision2      INTEGER;
        v_temp           INTEGER;
        v_base_version   INTEGER;
        v_revision_table TEXT;
    BEGIN
        IF NOT @extschema@.ver_is_table_versioned(%schema_name%, %table_name%) THEN
            RAISE EXCEPTION 'Table %full_table_name% is not versioned';
        END IF;
        
        v_revision1 := p_revision1;
        v_revision2 := p_revision2;
        IF v_revision1 = v_revision2 THEN
            RETURN;
        END IF;
        
        IF v_revision1 > v_revision2 THEN
            RAISE EXCEPTION 'Revision 1 (%) is greater than revision 2 (%)', v_revision1, v_revision2;
        END IF;
        
        SELECT @extschema@.ver_get_table_base_revision(%schema_name%, %table_name%)
        INTO   v_base_version;
        IF v_base_version > v_revision2 THEN
            RETURN;
        END IF;
        
        RETURN QUERY EXECUTE
        @extschema@.ver_ExpandTemplate(
            $sql$
            WITH changed_within_range AS (
                SELECT 
                    T.%key_col%,
                    min(T._revision_created) <= %1% AS existed,
                    max(T._revision_created) AS last_update_revision
                FROM
                    %revision_table% AS T
                WHERE (
                    (T._revision_created <= %1% AND T._revision_expired > %1% AND T._revision_expired <= %2%) OR
                    (T._revision_created > %1%  AND T._revision_created <= %2%)
                )
                GROUP BY
                    T.%key_col%
            )
            SELECT
                CAST(
                    CASE
                       WHEN TL._revision_expired <= %2% THEN 'D'
                       WHEN C.existed THEN 'U'
                       ELSE 'I'
                    END
                AS CHAR(1)) AS action,
%select_columns%
            FROM
                changed_within_range C
                JOIN %revision_table% AS TL ON TL.%key_col% = C.%key_col% AND TL._revision_created = C.last_update_revision
            WHERE
                C.existed OR
                (TL._revision_expired IS NULL OR TL._revision_expired > %2%);
            $sql$,
            ARRAY[
                v_revision1::TEXT,
                v_revision2::TEXT
            ]
        );
        RETURN;
    END;
$FUNC$ LANGUAGE plpgsql;

    $template$;
    
    v_sql := REPLACE(v_sql, '%func_sig%',       @extschema@._ver_get_diff_function(p_schema, p_table));
    v_sql := REPLACE(v_sql, '%table_columns%',  v_table_columns);
    v_sql := REPLACE(v_sql, '%schema_name%',    quote_literal(p_schema));
    v_sql := REPLACE(v_sql, '%table_name%',     quote_literal(p_table));
    v_sql := REPLACE(v_sql, '%full_table_name%', quote_ident(p_schema) || '.' || quote_ident(p_table));
    v_sql := REPLACE(v_sql, '%select_columns%', v_select_columns_diff);
    v_sql := REPLACE(v_sql, '%key_col%',        quote_ident(p_key_col));
    v_sql := REPLACE(v_sql, '%revision_table%', v_revision_table);
    
    EXECUTE 'DROP FUNCTION IF EXISTS ' || @extschema@._ver_get_diff_function(p_schema, p_table);
    EXECUTE v_sql;

    -- Create get version function for table called: 
    -- ver_get_$schema$_$table$_revision(p_revision integer)
    v_sql := $template$
    
CREATE OR REPLACE FUNCTION %func_sig%
RETURNS TABLE(
    %table_columns%
) AS
$FUNC$
BEGIN
    RETURN QUERY EXECUTE
    @extschema@.ver_ExpandTemplate(
        $sql$
            SELECT
%select_columns%
            FROM
                %revision_table% AS T
            WHERE
                _revision_created <= %1% AND
                (_revision_expired > %1% OR _revision_expired IS NULL)
        $sql$,
        ARRAY[
            p_revision::TEXT
        ]
    );
END;
$FUNC$ LANGUAGE plpgsql;

    $template$;
    
    v_sql := REPLACE(v_sql, '%func_sig%', @extschema@._ver_get_revision_function(p_schema, p_table));
    v_sql := REPLACE(v_sql, '%table_columns%', v_table_columns);
    v_sql := REPLACE(v_sql, '%select_columns%', v_select_columns_rev);
    v_sql := REPLACE(v_sql, '%revision_table%', v_revision_table);
    
    EXECUTE 'DROP FUNCTION IF EXISTS ' || @extschema@._ver_get_revision_function(p_schema, p_table);
    EXECUTE v_sql;

    RETURN TRUE;
END;
$$ LANGUAGE plpgsql;

/**
* Creates trigger and trigger function required for versioning the table.
*
* @param p_table_oid      The table regclass
* @param p_key_col        The unique non-compostite integer column key.
* @return                 If creating the functions was successful.
* @throws RAISE_EXCEPTION If the table is not versioned
*
* -- {
*/
CREATE OR REPLACE FUNCTION ver_create_version_trigger(
    p_table_oid   REGCLASS,
    p_key_col NAME
) 
RETURNS BOOLEAN AS
$$
DECLARE
    v_revision_table TEXT;
    v_sql            TEXT;
    v_trigger_name   VARCHAR;
    v_column_name    NAME;
    v_column_update  TEXT;
    v_col_insert_col TEXT;
    v_col_insert_val TEXT;
    v_table          NAME;
    v_schema         NAME;
    v_owner          NAME;
BEGIN

    SELECT nspname, relname, rolname
    INTO v_schema, v_table, v_owner
    FROM @extschema@._ver_get_table_info(p_table_oid);

    -- Check that SESSION_USER is the owner of the table, or
    -- refuse to add columns to it
    IF NOT pg_has_role(session_user, v_owner, 'usage') THEN
        RAISE EXCEPTION 'User % cannot create version triggers on table %'
            ' for lack of usage privileges on table owner role %',
            session_user, p_table_oid, v_owner;
    END IF;

    IF NOT @extschema@._ver_is_table_versioned(v_schema, v_table) THEN
        RAISE EXCEPTION 'Table %.% is not versioned', quote_ident(v_schema), quote_ident(v_table);
    END IF;
    
    v_revision_table := @extschema@.ver_get_version_table_full(v_schema, v_table);
    

    SELECT string_agg(quote_ident(att_name), ',') INTO v_col_insert_col
    FROM unnest(@extschema@._ver_get_table_columns(v_schema || '.' ||  v_table));

    SELECT string_agg('NEW.' || quote_ident(att_name), E',\n') INTO v_col_insert_val
    FROM unnest(@extschema@._ver_get_table_columns(v_schema || '.' ||  v_table));

    v_column_update := '';
    FOR v_column_name IN
        SELECT att_name AS column_name
        FROM unnest(@extschema@._ver_get_table_columns(v_schema || '.' ||  v_table))
    LOOP
        IF v_column_name = p_key_col THEN
            CONTINUE;
        END IF;
        IF v_column_update != '' THEN
            v_column_update := v_column_update || E',\n                        ';
        END IF;
        
        v_column_update := v_column_update || quote_ident(v_column_name) || ' = NEW.' 
            || quote_ident(v_column_name);
    END LOOP;
    
    v_sql := $template$

CREATE OR REPLACE FUNCTION %revision_table%() RETURNS trigger AS $TRIGGER$
    DECLARE
       v_revision      @extschema@.revision.id%TYPE;
       v_last_revision @extschema@.revision.id%TYPE;
       v_table_id      @extschema@.versioned_tables.id%TYPE;
    BEGIN
        BEGIN
            SELECT
                max(VER.revision)
            INTO
                v_revision
            FROM
                _changeset_revision VER;
                
            IF v_revision IS NULL THEN
                RAISE EXCEPTION 'Versioning system information is missing';
            END IF;
        EXCEPTION
            WHEN undefined_table THEN
                RAISE EXCEPTION 'To begin editing %full_table_name% you need to create a revision';
        END;

        SELECT
            VTB.id
        INTO
            v_table_id
        FROM
            @extschema@.versioned_tables VTB
        WHERE
            VTB.table_name = %table_name% AND
            VTB.schema_name = %schema_name%;
        
        IF v_table_id IS NULL THEN
            RAISE EXCEPTION 'Table versioning system information is missing for %full_table_name%';
        END IF;

        IF NOT EXISTS (
            SELECT TRUE
            FROM   @extschema@.tables_changed
            WHERE  table_id = v_table_id
            AND    revision = v_revision
        )
        THEN
            INSERT INTO @extschema@.tables_changed(revision, table_id)
            VALUES (v_revision, v_table_id);
        END IF;

        
        IF (TG_OP <> 'INSERT') THEN
            SELECT 
                _revision_created INTO v_last_revision
            FROM 
                %revision_table%
            WHERE 
                %key_col% = OLD.%key_col% AND
                _revision_expired IS NULL;

            IF v_last_revision = v_revision THEN
                IF TG_OP = 'UPDATE' AND OLD.%key_col% = NEW.%key_col% THEN
                    UPDATE
                        %revision_table%
                    SET
                        %revision_update_cols%
                    WHERE
                        %key_col% = NEW.%key_col% AND
                        _revision_created = v_revision AND
                        _revision_expired IS NULL;
                    RETURN NEW;
                ELSE
                    DELETE FROM 
                        %revision_table%
                    WHERE
                        %key_col% = OLD.%key_col% AND
                        _revision_created = v_last_revision;
                END IF;
            ELSE
                UPDATE
                    %revision_table%
                SET
                    _revision_expired = v_revision
                WHERE
                    %key_col% = OLD.%key_col% AND
                    _revision_created = v_last_revision;
            END IF;
        END IF;

        IF( TG_OP <> 'DELETE') THEN
            INSERT INTO %revision_table% (
                _revision_created,
                _revision_expired,
                %revision_insert_cols%
            )
            SELECT
                v_revision,
                NULL,
                %revision_insert_vals%;
            RETURN NEW;
        END IF;
        
        RETURN NULL;
    END;
$TRIGGER$ LANGUAGE plpgsql SECURITY DEFINER;

    $template$;

    v_sql := REPLACE(v_sql, '%schema_name%',    quote_literal(v_schema));
    v_sql := REPLACE(v_sql, '%table_name%',     quote_literal(v_table));
    v_sql := REPLACE(v_sql, '%full_table_name%', quote_ident(v_schema) || '.' || quote_ident(v_table));
    v_sql := REPLACE(v_sql, '%key_col%',        quote_ident(p_key_col));
    v_sql := REPLACE(v_sql, '%revision_table%', v_revision_table);
    v_sql := REPLACE(v_sql, '%revision_update_cols%', v_column_update);
    v_sql := REPLACE(v_sql, '%revision_insert_cols%', v_col_insert_col);
    v_sql := REPLACE(v_sql, '%revision_insert_vals%', v_col_insert_val);
    
    EXECUTE v_sql;

    SELECT @extschema@._ver_get_version_trigger(v_schema, v_table)
    INTO v_trigger_name;
    

    EXECUTE 'DROP TRIGGER IF EXISTS '  || v_trigger_name || ' ON ' ||  
        quote_ident(v_schema) || '.' || quote_ident(v_table);

    EXECUTE 'CREATE TRIGGER '  || v_trigger_name || ' AFTER INSERT OR UPDATE OR DELETE ON ' ||  
        quote_ident(v_schema) || '.' || quote_ident(v_table) ||
        ' FOR EACH ROW EXECUTE PROCEDURE ' || v_revision_table || '()';
    
    RETURN TRUE;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;
--}

/*
* Creates trigger and trigger function required for versioning the table.
*
* @param p_schema         The table schema
* @param p_table          The table name
* @param p_key_col        The unique non-compostite integer column key.
* @return                 If creating the functions was successful.
* @throws RAISE_EXCEPTION If the table is not versioned
*
* {
*/
CREATE OR REPLACE FUNCTION ver_create_version_trigger(
    p_schema  NAME,
    p_table   NAME,
    p_key_col NAME
) 
RETURNS BOOLEAN AS
$$
    SELECT @extschema@.ver_create_version_trigger(
        ( p_schema || '.' || p_table)::regclass,
        p_key_col
    );
$$ LANGUAGE sql;
-- }

--
-- Fix existing table triggers if coming from a version prior to 1.5.0
--
-- WARNING: this needs to be sourced _before_ the script defining
--          ver_version() function
-- {
DO $$
DECLARE
    old_version TEXT;
    is_extension BOOL;
    rec RECORD;
BEGIN
    BEGIN
        old_version := @extschema@.ver_version();
    EXCEPTION WHEN undefined_function THEN
        RAISE DEBUG 'ver_version not available, '
                    'we are either doing a new install '
                    'or coming from version before 1.3';
    END;

    SELECT EXISTS ( SELECT * FROM pg_catalog.pg_extension
                    WHERE extname = 'table_version' )
    INTO is_extension;

    -- We only need to update triggers when coming
    -- from versions 1.4 or earlier. Function ver_version
    -- was introduced in 1.3 so no need to check any previous
    -- version (old_version would be null in those cases)
    --
    -- Note that if old_version would also be NULL for new
    -- installs in which case the SELECT below will do nothing
    -- as ver_get_versioned_tables() would return the empty set
    --
    IF ( old_version IS NULL OR
       old_version LIKE '1.3.%' OR
       old_version LIKE '1.4.%' ) AND
       EXISTS (
          SELECT * FROM @extschema@.ver_get_versioned_tables()
       ) versioned_tables
    THEN
        RAISE NOTICE 'Updating triggers on versioned tables';
        FOR rec IN SELECT schema_name, table_name,
            @extschema@.ver_create_version_trigger(
                schema_name, table_name, key_column)
            FROM (
                SELECT * FROM @extschema@.ver_get_versioned_tables()
            ) versioned_tables
        LOOP
            IF is_extension THEN
                EXECUTE format('ALTER EXTENSION table_version '
                    ' DROP FUNCTION '
                    '@extschema@.%s_%s_revision()',
                    rec.schema_name, rec.table_name);
            END IF;
        END LOOP;
    END IF;
END
$$ LANGUAGE 'plpgsql';
--}
/**
* Processes a text template given a set of input template parameters. Template 
* parameters within the text are substituted content must be written as '%1%' 
* to '%n%' where n is the number of text parameters.
*
* @param p_template       The template text
* @param p_params         The template parameters
* @return                 The expanded template text
*/
CREATE OR REPLACE FUNCTION ver_ExpandTemplate (
    p_template TEXT,
    p_params TEXT[]
)
RETURNS
    TEXT AS
$$
DECLARE 
    v_expanded TEXT;
BEGIN
    v_expanded := p_template;
    FOR i IN 1 .. array_length(p_params,1) LOOP
        v_expanded := REPLACE( v_expanded, '%' || i || '%', p_params[i]);
    END LOOP;
    RETURN v_expanded;
END;
$$
LANGUAGE plpgsql;

/**
* Executes a text template given a set of input template parameters. Template 
* parameters within the text are substituted content must be written as '%1%' 
* to '%n%' where n is the number of text parameters.
*
* @param p_template       The template text
* @param p_params         The template parameters
* @return                 The number of rows affected by running the template
*/

CREATE OR REPLACE FUNCTION ver_ExecuteTemplate(
    p_template TEXT,
    p_params TEXT[])
RETURNS
    BIGINT AS
$$
DECLARE
    v_sql TEXT;
    v_count BIGINT;
BEGIN
    v_sql := @extschema@.ver_ExpandTemplate( p_template, p_params );
    BEGIN
        EXECUTE v_sql;
    EXCEPTION
        WHEN others THEN
            RAISE EXCEPTION E'Error executing template SQL: %\nError: %',
            v_sql, SQLERRM;
    END;
    GET DIAGNOSTICS v_count=ROW_COUNT;
    RETURN v_count;
END;
$$
LANGUAGE plpgsql;
  
/**
* Gets the tablename for the tables revision data.
*
* @param p_schema         The table schema
* @param p_table          The table name
* @return                 The revision data table name
*/
CREATE OR REPLACE FUNCTION ver_get_version_table(
    p_schema NAME,
    p_table NAME
) 
RETURNS VARCHAR AS $$
    SELECT quote_ident($1 || '_' || $2 || '_revision');
$$ LANGUAGE sql IMMUTABLE;

/**
* Gets the fully qualified tablename for the tables revision data.
*
* @param p_schema         The table schema
* @param p_table          The table name
* @return                 The revision data fully qualified table name
*/
CREATE OR REPLACE FUNCTION ver_get_version_table_full(
    p_schema NAME,
    p_table NAME
) 
RETURNS VARCHAR AS $$
    SELECT '@extschema@.' || @extschema@.ver_get_version_table($1, $2);
$$ LANGUAGE sql IMMUTABLE;

/**
* Gets the trigger name that is created on the versioned table.
*
* @param p_schema         The table schema
* @param p_table          The table name
* @return                 The trigger name
*/
CREATE OR REPLACE FUNCTION _ver_get_version_trigger(
    p_schema NAME,
    p_table NAME
)
RETURNS VARCHAR AS $$
    SELECT quote_ident($1 || '_' || $2 || '_revision_trg');
$$ LANGUAGE sql IMMUTABLE;

/**
* Gets the changset difference function name and signature that is created for the versioned table.
*
* @param p_schema         The table schema
* @param p_table          The table name
* @return                 The function name and signature
*/
CREATE OR REPLACE FUNCTION _ver_get_diff_function(
    p_schema NAME,
    p_table NAME
)
RETURNS VARCHAR AS $$
    SELECT ('@extschema@.' || quote_ident('ver_get_' || $1 || '_' || $2 || '_diff') || '(p_revision1 INTEGER, p_revision2 INTEGER)');
$$ LANGUAGE sql IMMUTABLE;

/**
* Gets the revision function name and signature that is created for the versioned table.
*
* @param p_schema         The table schema
* @param p_table          The table name
* @return                 The function name and signature
*/
CREATE OR REPLACE FUNCTION _ver_get_revision_function(
    p_schema NAME,
    p_table NAME
)
RETURNS VARCHAR AS $$
    SELECT ('@extschema@.' || quote_ident('ver_get_' || $1 || '_' || $2 || '_revision') || '(p_revision INTEGER)');
$$ LANGUAGE sql IMMUTABLE;

/**
* Determine if a temp table exists within the current SQL session.
*
* @param p_table_name     The name of the temp table
* @return                 If true if the table exists.
*/
CREATE OR REPLACE FUNCTION _ver_get_reversion_temp_table(
    p_table_name NAME
)
RETURNS BOOLEAN AS
$$
DECLARE
    v_exists BOOLEAN;
BEGIN
    SELECT
        TRUE
    INTO
        v_exists
    FROM
        pg_catalog.pg_class c
        LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
    WHERE
        n.nspname LIKE 'pg_temp_%' AND
        pg_catalog.pg_table_is_visible(c.oid) AND
        c.relkind = 'r' AND
        c.relname = p_table_name;

    IF v_exists IS NULL THEN
        v_exists := FALSE;
    END IF;

    RETURN v_exists;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;

/**
* Get the owner for a table
*
* @param p_table          The table
* @return                 Table owner rolename
*/
CREATE OR REPLACE FUNCTION _ver_get_table_owner(
    p_table REGCLASS
)
RETURNS TEXT AS
$$
    SELECT  quote_ident(r.rolname)
    FROM   pg_catalog.pg_class c
    JOIN   pg_catalog.pg_roles r on (c.relowner = r.oid)
    WHERE  c.oid = p_table
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_table_key_datatype(
    p_table      REGCLASS,
    p_key_column NAME
)
RETURNS TEXT AS
$$
    SELECT
        format_type(ATT.atttypid, NULL)
    FROM
        pg_index IDX,
        pg_attribute ATT
    WHERE
        IDX.indrelid = $1 AND
        ATT.attrelid = $1 AND
        ATT.attnum = ANY(IDX.indkey) AND
        ATT.attnotnull = TRUE AND
        IDX.indisunique = TRUE AND
        IDX.indexprs IS NULL AND
        IDX.indpred IS NULL AND
        array_length(IDX.indkey::INTEGER[], 1) = 1 AND
        LOWER(ATT.attname) = LOWER($2)
    ORDER BY
        IDX.indisprimary DESC;
$$ LANGUAGE sql;

CREATE OR REPLACE FUNCTION ver_table_key_is_valid(
    p_table      REGCLASS,
    p_key_column NAME
)
RETURNS BOOLEAN AS
$$
    SELECT EXISTS (
        SELECT
            TRUE
        FROM
            pg_index IDX,
            pg_attribute ATT
        WHERE
            IDX.indrelid = $1 AND
            ATT.attrelid = $1 AND
            ATT.attnum = ANY(IDX.indkey) AND
            ATT.attnotnull = TRUE AND
            IDX.indisunique = TRUE AND
            IDX.indexprs IS NULL AND
            IDX.indpred IS NULL AND
            format_type(ATT.atttypid, NULL) IN (
                'integer', 'bigint', 'text', 'character varying'
            ) AND
            array_length(IDX.indkey::INTEGER[], 1) = 1 AND
            LOWER(ATT.attname) = LOWER($2)
        ORDER BY
            IDX.indisprimary DESC
    );
$$ LANGUAGE sql;

--------------------------------------------------------------------------------

-- postgresql-table_version - PostgreSQL database patch change management extension
--
-- Copyright 2016 Crown copyright (c)
-- Land Information New Zealand and the New Zealand Government.
-- All rights reserved
--
-- This software is released under the terms of the new BSD license. See the 
-- LICENSE file for more information.
--
--------------------------------------------------------------------------------

DO $$ BEGIN IF NOT EXISTS (
  SELECT t.oid
  FROM pg_type t, pg_namespace n
  WHERE n.oid = t.typnamespace
    AND n.nspname = '@extschema@'
    AND t.typname = 'attribute'
) THEN
  CREATE TYPE ATTRIBUTE AS (
      att_name NAME,
      att_type NAME,
      att_not_null BOOLEAN
  );
END IF;
END; $$;

CREATE OR REPLACE FUNCTION ver_get_table_differences(
    p_table1      REGCLASS,
    p_table2      REGCLASS,
    p_compare_key NAME
)
RETURNS SETOF RECORD
AS $$
DECLARE
    v_table_1_cols  @extschema@.ATTRIBUTE[];
    v_table_1_uniq  @extschema@.ATTRIBUTE[];
    v_table_2_cols  @extschema@.ATTRIBUTE[];
    v_common_cols   @extschema@.ATTRIBUTE[];
    v_unique_cols   @extschema@.ATTRIBUTE[];
    v_sql           TEXT;
    v_error         TEXT;
    v_geom_col      TEXT;
BEGIN
    IF p_table1 = p_table2 THEN
        RETURN;
    END IF;

    v_sql := '';

    IF NOT @extschema@.ver_table_key_is_valid(p_table1, p_compare_key) THEN
        RAISE EXCEPTION
            '''%'' is not a unique non-composite integer, bigint, text, or varchar column for %',
            p_compare_key, CAST(p_table1 AS TEXT);
    END IF;

    IF NOT @extschema@.ver_table_key_is_valid(p_table2, p_compare_key) THEN
        RAISE EXCEPTION
            '''%'' is not a  unique non-composite integer, bigint, text, or varchar column for %',
            p_compare_key, CAST(p_table2 AS TEXT);
    END IF;
    
    SELECT @extschema@._ver_get_table_columns(p_table1)
    INTO v_table_1_cols;
    
    SELECT @extschema@._ver_get_table_columns(p_table2)
    INTO v_table_2_cols;
    
    SELECT @extschema@._ver_get_table_unique_constraint_columns(p_table1)
    INTO v_table_1_uniq;

    SELECT ARRAY(
        SELECT ROW(ATT.att_name, ATT.att_type, ATT.att_not_null)
        FROM   unnest(v_table_1_cols) AS ATT 
        WHERE  ATT.att_name IN 
            (SELECT (unnest(v_table_2_cols)).att_name)
    )
    INTO v_common_cols;

    SELECT ARRAY(
        SELECT ROW(ATT.att_name, ATT.att_type, ATT.att_not_null)
        FROM   unnest(v_table_1_cols) AS ATT 
        WHERE  ATT.att_name IN 
            (SELECT (unnest(v_table_2_cols)).att_name)
        AND ATT.att_name IN
            (SELECT (unnest(v_table_1_uniq)).att_name)
        AND ATT.att_name <> p_compare_key
    )
    INTO v_unique_cols;
    
    v_sql := table_version.ver_ExpandTemplate( $sql$
	SELECT
	    CASE WHEN t2.%1% IS NULL THEN
		    CAST('D' AS CHAR(1))
	    WHEN t1.%1% IS NULL THEN
		    CAST('I' AS CHAR(1))
	    WHEN CAST(%2% AS TEXT) <> CAST(%3% AS TEXT) THEN
		    CAST('X' AS CHAR(1))
	    ELSE
		    CAST('U' AS CHAR(1))
	    END AS action,
	    CASE WHEN t2.%1% IS NULL THEN
		    t1.%1%
	    ELSE
		    t2.%1%
	    END AS id
	FROM
	    (SELECT %4% FROM %5%) AS t1
	    FULL OUTER JOIN
	    (SELECT %4% FROM %6%) AS t2
	    ON t2.%1% = t1.%1%
	WHERE
	    t1.%1% IS NULL OR
	    t2.%1% IS NULL OR
	    CAST(%2% AS TEXT) <> CAST(%3% AS TEXT) OR
	    NOT COALESCE(t1.* = t2.*, FALSE)
        $sql$,
        ARRAY[
            quote_ident(p_compare_key),
            @extschema@._ver_get_compare_sql(v_unique_cols,'t1'),
            @extschema@._ver_get_compare_sql(v_unique_cols,'t2'),
            (SELECT array_to_string(array_agg(quote_ident(att_name)), ',')
              att_name FROM unnest(v_common_cols)),
            p_table1::TEXT,
            p_table2::TEXT
        ]
    );
    
    FOR v_geom_col IN
        SELECT att_name
        FROM   unnest(v_common_cols)
        WHERE  att_type LIKE 'geometry%'
    LOOP
        -- binary compare of PostGIS geometry including SRID (faster than text compare)
        v_sql := v_sql || ' OR ST_AsEWKB(t1.' || v_geom_col || ') <> ST_AsEWKB(t2.' || v_geom_col || ')';
    END LOOP;

    RAISE DEBUG 'diff sql = %', v_sql;

    RETURN QUERY EXECUTE v_sql;
EXCEPTION
    WHEN others THEN
        GET STACKED DIAGNOSTICS v_error = PG_EXCEPTION_CONTEXT;
        RAISE EXCEPTION E'Failed comparing tables\n%\nERROR: %', v_error, SQLERRM;
END;
$$ LANGUAGE plpgsql;
    
CREATE OR REPLACE FUNCTION ver_apply_table_differences(
    p_original_table   REGCLASS,
    p_new_table        REGCLASS,
    p_key_column       NAME,
    OUT number_inserts BIGINT,
    OUT number_deletes BIGINT,
    OUT number_updates BIGINT
)
AS $$
DECLARE
    v_nuniqf  BIGINT DEFAULT 0;
BEGIN
    
    RAISE NOTICE 'Generating difference data for %', p_original_table;
    
    PERFORM @extschema@.ver_ExecuteTemplate( $sql$
        CREATE TEMP TABLE table_diff AS
        SELECT
            T.id,
            T.action
        FROM
            @extschema@.ver_get_table_differences(
                '%1%', '%2%', '%3%'
            ) AS T (action CHAR(1), id %4%)
        ORDER BY
            T.action,
            T.id;
        $sql$,
        ARRAY[
            p_original_table::TEXT,
            p_new_table::TEXT,
            p_key_column::TEXT,
            @extschema@.ver_table_key_datatype(p_original_table, p_key_column)
        ]
    );
    
    RAISE NOTICE 'Completed generating difference data for %', p_original_table;
    
    ALTER TABLE table_diff ADD PRIMARY KEY (id);
    ANALYSE table_diff;
    
    SELECT * FROM @extschema@._ver_apply_changes(
        p_original_table, p_new_table, 'table_diff', p_key_column
    )
    INTO number_inserts, number_deletes, number_updates;

    DROP TABLE table_diff;
    
    RETURN;
END;
$$ LANGUAGE plpgsql;
            
            
CREATE OR REPLACE FUNCTION _ver_apply_changes(
    p_original_table REGCLASS,
    p_new_table REGCLASS,
    p_diff_table NAME,
    p_key_column NAME,
    OUT number_inserts BIGINT,
    OUT number_deletes BIGINT,
    OUT number_updates BIGINT
)
    AS $$
DECLARE
    v_nuniqf  BIGINT DEFAULT 0;
    v_count   BIGINT;
BEGIN

    number_inserts := 0;
    number_deletes := 0;
    number_updates := 0;
    
    EXECUTE 'SELECT 1 FROM ' || p_diff_table || ' LIMIT 1'
    INTO v_count;
   
    IF v_count THEN
        EXECUTE 'SELECT count(*) FROM ' || p_diff_table || ' WHERE action=' || quote_literal('X')
        INTO v_nuniqf;
        
        IF v_nuniqf > 0 THEN
            RAISE NOTICE
                '% updates changed to delete/insert in % to avoid potential uniqueness constraint errors',
                v_nuniqf, p_original_table;
        END IF;
        
        RAISE NOTICE 'Deleting from % using difference data', p_original_table;
        
        number_deletes := @extschema@._ver_apply_inc_delete(
            p_original_table, p_diff_table, p_key_column
        );
        
        RAISE NOTICE 'Updating % using difference data', p_original_table;
        
        number_updates :=  @extschema@._ver_apply_inc_update(
            p_original_table, p_diff_table, p_new_table, p_key_column
        );
        
        RAISE NOTICE 'Inserting into % using difference data', p_original_table;
        
        number_inserts := @extschema@._ver_apply_inc_insert(
            p_original_table, p_diff_table, p_new_table, p_key_column
        );
        
        RAISE NOTICE 'Finished updating % using difference data', p_original_table;

        number_deletes := number_deletes - v_nuniqf;
        number_inserts := number_inserts - v_nuniqf;
        number_updates := number_updates + v_nuniqf;
    END IF;
    
    RETURN;
END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION _ver_apply_inc_delete(
    p_delete_table REGCLASS,
    p_inc_change_table NAME,
    p_key_column NAME
)
RETURNS
    BIGINT
AS
$$
BEGIN
    RETURN @extschema@.ver_ExecuteTemplate( $sql$
        DELETE FROM %1% AS T
        USING %2% AS INC
        WHERE T.%3% = INC.id
        AND  INC.action IN ('D','X')
        $sql$,
        ARRAY[p_delete_table::text,p_inc_change_table,quote_ident(p_key_column)]
    );
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION _ver_apply_inc_update(
    p_update_table REGCLASS,
    p_inc_change_table NAME,
    p_inc_data_table REGCLASS,
    p_key_column NAME
)
RETURNS
    BIGINT
AS
$$
DECLARE
    v_sql TEXT;
    v_update_col_txt TEXT;
    v_table_cols @extschema@.ATTRIBUTE[];
    v_col @extschema@.ATTRIBUTE;
BEGIN
    v_table_cols := @extschema@._ver_get_table_columns(p_update_table);
    IF v_table_cols IS NULL THEN
        RAISE EXCEPTION 'Could not find any table columns for %',
            p_update_table;
    END IF;
    
    v_update_col_txt := '';
    FOR v_col IN SELECT * FROM unnest(v_table_cols) LOOP
        IF v_update_col_txt != '' THEN
            v_update_col_txt := v_update_col_txt || ',';
        END IF;
        v_update_col_txt := v_update_col_txt || quote_ident(v_col.att_name) ||
            ' = NEW_DAT.' || quote_ident(v_col.att_name);
    END LOOP;
    
    RETURN @extschema@.ver_ExecuteTemplate( $sql$
        UPDATE %1% AS CUR
        SET %2%
        FROM %3% AS NEW_DAT,
             %4% AS INC
        WHERE INC.id = CUR.%5% 
        AND   NEW_DAT.%5% = CUR.%5%
        AND   INC.action = 'U'
        $sql$,
        ARRAY[
            p_update_table::text,
            v_update_col_txt,
            p_inc_data_table::text,
            p_inc_change_table,
            quote_ident(p_key_column)
        ]
    );
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION _ver_apply_inc_insert(
    p_insert_table REGCLASS,
    p_inc_change_table NAME,
    p_inc_data_table REGCLASS,
    p_key_column NAME
)
RETURNS
    BIGINT
AS
$$
DECLARE
    v_table_cols text;
BEGIN
    SELECT array_to_string(array_agg(quote_ident(att_name)), ',') 
    INTO v_table_cols
    FROM unnest(@extschema@._ver_get_table_columns(p_insert_table));
        
    IF v_table_cols = '' THEN
        RAISE EXCEPTION 'Could not find any table columns for %',
            p_insert_table;
    END IF;
    
    RETURN @extschema@.ver_ExecuteTemplate( $sql$
        INSERT INTO %1% (%2%)
        SELECT %2% FROM %3%
        WHERE %4% IN
          (SELECT id FROM %5% WHERE action IN ('I','X'))
        $sql$,
        ARRAY[
            p_insert_table::text,
            v_table_cols,
            p_inc_data_table::text,
            quote_ident(p_key_column),
            p_inc_change_table
        ]
    );
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION _ver_get_compare_sql(
    p_columns     @extschema@.ATTRIBUTE[],
    p_table_alias TEXT
)
RETURNS TEXT AS 
$$
DECLARE
    v_sql          TEXT;
    v_col_name     NAME;
    v_col_type     TEXT;
    v_col_not_null BOOLEAN;
BEGIN
    IF array_ndims(p_columns) IS NULL THEN
        RETURN quote_literal('');
    END IF;
    v_sql := '';
    FOR v_col_name, v_col_type, v_col_not_null IN
        SELECT
            att_name,
            att_type,
            att_not_null
        FROM
            unnest(p_columns)
        ORDER BY
            att_name,
            att_type,
            att_not_null
    LOOP
        IF v_sql != '' THEN
            v_sql := v_sql || ' || ';
        END IF;

        IF v_col_not_null THEN
            v_sql := v_sql || '''|V'' || ' || 'CAST(' || p_table_alias ||
                '.' || quote_ident(v_col_name) || ' AS TEXT)';
        ELSE
            v_sql := v_sql || 'COALESCE(''V|'' || CAST(' || p_table_alias ||
                '.' || quote_ident(v_col_name) || ' AS TEXT), ''|N'')';
        END IF;
    END LOOP;

    RETURN v_sql;
END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION _ver_get_table_unique_constraint_columns(
    p_table REGCLASS,
    p_key_column NAME = NULL,
    p_return_comp_keys BOOLEAN = TRUE
)
RETURNS @extschema@.ATTRIBUTE[] AS
$$
    SELECT array_agg(
        CAST((attname, type, attnotnull) AS @extschema@.ATTRIBUTE)
    )
    FROM
        (
        SELECT
            ATT.attname,
            format_type(ATT.atttypid, ATT.atttypmod) as type,
            ATT.attnotnull
        FROM
            pg_index IDX,
            pg_attribute ATT
        WHERE
            ATT.attrelid = $1 AND
            (p_key_column IS NULL OR ATT.attname <> $2) AND
            IDX.indrelid = ATT.attrelid AND
            IDX.indisunique = TRUE AND
            IDX.indexprs IS NULL AND
            IDX.indpred IS NULL AND
            ATT.attnum IN (
                SELECT IDX.indkey[i]
                FROM   generate_series(0, IDX.indnatts) AS i
                WHERE  ($3 OR array_length(IDX.indkey,1) = 1)
            )
        UNION
        SELECT
            ATT.attname,
            format_type(ATT.atttypid, ATT.atttypmod) as type,
            ATT.attnotnull
        FROM
            pg_attribute ATT
        WHERE
            ATT.attnum > 0 AND
            (p_key_column IS NULL OR ATT.attname <> $2) AND
            NOT ATT.attisdropped AND
            ATT.attrelid = $1 AND
            ATT.attnum IN
            (SELECT unnest(conkey)
             FROM pg_constraint
             WHERE 
                conrelid = p_table AND
                contype in ('p','u') AND
                ($3 OR array_length(conkey,1) = 1))
    ) AS ATT
$$ LANGUAGE sql;

-- Return a list of columns for a table as an array of ATTRIBUTE entries

CREATE OR REPLACE FUNCTION _ver_get_table_columns(
    p_table REGCLASS
)
RETURNS @extschema@.ATTRIBUTE[] AS
$$
    SELECT array_agg(
        CAST((ATT.attname, format_type(ATT.atttypid, ATT.atttypmod), 
         ATT.attnotnull) AS @extschema@.ATTRIBUTE))
    FROM
        pg_attribute ATT
    WHERE
        ATT.attnum > 0 AND
        NOT ATT.attisdropped AND
        ATT.attrelid = p_table;
$$ LANGUAGE sql;

-- {
CREATE OR REPLACE FUNCTION ver_versioned_table_drop_column(
    p_table_oid   REGCLASS,
    p_column_name NAME
)
RETURNS BOOLEAN AS
$$
DECLARE
    v_key_col NAME;
    v_trigger_name TEXT;
    v_revision_table TEXT;
    v_schema          NAME;
    v_table           NAME;
    v_owner           NAME;
BEGIN

    SELECT nspname, relname, rolname
    INTO v_schema, v_table, v_owner
    FROM @extschema@._ver_get_table_info(p_table_oid);

    -- Check that SESSION_USER is the owner of the table, or
    -- refuse to add columns to it
    IF NOT pg_has_role(session_user, v_owner, 'usage') THEN
        RAISE EXCEPTION 'User % cannot drop columns from table %'
            ' for lack of usage privileges on table owner role %',
            session_user, p_table_oid, v_owner;
    END IF;

    IF NOT @extschema@.ver_is_table_versioned(v_schema, v_table) THEN
        RAISE EXCEPTION 'Table %.% is not versioned', quote_ident(v_schema), quote_ident(v_table);
    END IF;

    v_revision_table := @extschema@.ver_get_version_table_full(v_schema, v_table);
    v_trigger_name := @extschema@._ver_get_version_trigger(v_schema, v_table);

    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' || quote_ident(v_table) ||      ' DISABLE TRIGGER ' || v_trigger_name;
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' || quote_ident(v_table) ||      ' DROP COLUMN ' || quote_ident(p_column_name);
    EXECUTE 'ALTER TABLE ' || v_revision_table      || ' DROP COLUMN ' || quote_ident(p_column_name);
    EXECUTE 'ALTER TABLE ' || quote_ident(v_schema) || '.' || quote_ident(v_table) ||      ' ENABLE TRIGGER ' || v_trigger_name;
    
    SELECT
        key_column
    INTO
        v_key_col
    FROM 
        @extschema@.versioned_tables
    WHERE
        schema_name = v_schema AND
        table_name = v_table;
    
    PERFORM @extschema@.ver_create_table_functions(v_schema, v_table, v_key_col);
    PERFORM @extschema@.ver_create_version_trigger(p_table_oid, v_key_col);
    RETURN TRUE;
END;
$$ LANGUAGE plpgsql SECURITY DEFINER;
--}

CREATE OR REPLACE FUNCTION ver_versioned_table_drop_column(
    p_schema_name NAME,
    p_table_name  NAME,
    p_column_name NAME
)
RETURNS BOOLEAN AS
$$
    SELECT @extschema@.ver_versioned_table_drop_column(
        ( p_schema_name || '.' || p_table_name)::regclass,
        p_column_name
    );
$$ LANGUAGE sql;
--
-- Change ID of any revision having start_time at a later time than
-- any other revision with higher ID
--
-- New IDs for revisions to be moved will be assigned in
-- start_time order and start after the highest existing
-- revision ID.
--
-- The number of moved revisions is returned.
--
-- {
CREATE OR REPLACE FUNCTION ver_fix_revision_disorder()
RETURNS bigint AS
$FIX$
DECLARE

  v_rec RECORD;
  v_newid bigint;
  v_numversionedtables int;
  v_totdisordered bigint;
  v_numdisordered bigint;

BEGIN

-- 1. Make sure sequence is set to stop filling gaps

  PERFORM setval('@extschema@.revision_id_seq',
    greatest(
      (select max(id) FROM @extschema@.revision),
      (select last_value from @extschema@.revision_id_seq)
    ), true);

-- 2. Prepare update queries
  v_numversionedtables := 0;
  FOR v_rec IN SELECT schema_name, table_name
               FROM table_version.versioned_tables
  LOOP
    BEGIN
      EXECUTE format('PREPARE "p_uc%s" AS '
          'UPDATE table_version.%s_%s_revision '
          'SET _revision_created = $1 WHERE _revision_created = $2',
          v_numversionedtables, v_rec.schema_name, v_rec.table_name);
      EXECUTE format('PREPARE "p_ue%s" AS '
          'UPDATE table_version.%s_%s_revision '
          'SET _revision_expired = $1 WHERE _revision_expired = $2',
          v_numversionedtables, v_rec.schema_name, v_rec.table_name);
      v_numversionedtables := v_numversionedtables + 1;
    EXCEPTION WHEN UNDEFINED_TABLE THEN

      RAISE WARNING 'Spurious record in table_version.versioned_tables '
                    'for table % in schema %: % does not exist)',
                    v_rec.table_name, v_rec.schema_name,
                    format('table_version.%s_%s_revision',
                      v_rec.schema_name, v_rec.table_name);
    END;
  END LOOP;

-- 3. For each misplaced revision, move to correct place

  v_totdisordered := 0;

  LOOP

    v_numdisordered := 0;

    FOR v_rec IN
      WITH revs_by_id AS (
        SELECT
          row_number() OVER (ORDER BY id) seq,
          id,
          start_time
        FROM
          table_version.revision
      ),
      revs_by_time AS (
        SELECT
          row_number() OVER (ORDER BY start_time) seq,
          id,
          start_time
        FROM
          table_version.revision
      )
      SELECT
          a.id,
          a.start_time atm,
          b.id as bi,
          b.start_time btm
        FROM revs_by_id a, revs_by_time b
       WHERE a.seq = b.seq
         AND a.id < b.id
         AND a.start_time > b.start_time
       ORDER by a.start_time

    LOOP

      -- Revision v_rec.id has to be moved to nextval('@extschema@.revision_id_seq');

      v_numdisordered := v_numdisordered + 1;

      -- Create new revision v_record

      INSERT INTO @extschema@.revision
        (id, revision_time, start_time, user_name, schema_change, comment)
      SELECT
        nextval('@extschema@.revision_id_seq'::regclass),
        revision_time, start_time, user_name, schema_change,
        comment
      FROM @extschema@.revision
        WHERE id = v_rec.id
      RETURNING id
      INTO v_newid;

      RAISE NOTICE 'Revision % has start_time %, while % has start_time %: renamed % to %',
        v_rec.id, v_rec.atm, v_rec.bi, v_rec.btm, v_rec.id, v_newid;

      -- Update @extschema@.tables_changed

      UPDATE @extschema@.tables_changed
        SET revision = v_newid
        WHERE revision = v_rec.id;
        
      -- Update all revisions of all revisioned tables


      FOR v_i IN 0..v_numversionedtables-1
      LOOP
        EXECUTE format('EXECUTE "p_uc%s"(%s, %s)', v_i, v_newid, v_rec.id);
        EXECUTE format('EXECUTE "p_ue%s"(%s, %s)', v_i, v_newid, v_rec.id);
      END LOOP;

      -- Delete now hopefully unreferenced old revision

      DELETE FROM @extschema@.revision WHERE id = v_rec.id;

    END LOOP;

    v_totdisordered := v_totdisordered + v_numdisordered;

    IF v_numdisordered = 0 THEN
      EXIT;
    END IF;

  END LOOP;

-- 4. Deallocate prepared update queries
  FOR i IN 0..v_numversionedtables-1
  LOOP
    EXECUTE format('DEALLOCATE "p_uc%s"', i);
    EXECUTE format('DEALLOCATE "p_ue%s"', i);
  END LOOP;

  RETURN v_totdisordered;

END;
$FIX$
LANGUAGE 'plpgsql' VOLATILE;
--}

CREATE OR REPLACE FUNCTION ver_version()
RETURNS TEXT AS $$ SELECT '1.6.0 '::text $$ LANGUAGE sql;

GRANT USAGE ON SCHEMA table_version TO public;
COMMIT;
